#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <clapack.h>

/**
 Rutina que lee la matriz de entrada almacenada en un archivo
 Args: nombre_archivo referencia_a_la_matriz_de_salida
**/
int lee_archivo( char *nombrearchivo, double A[][5], int *mynn ) 
{
	char linea[80];
	FILE *arch;
	int i, j, k;
	double val;
	char *ptr;

	arch = fopen( nombrearchivo, "r" );
	if( arch == NULL ) {
		fprintf( stderr, "No se puede abrir archivo %s\n", nombrearchivo );
		return( -1 );
	}

	for( i=0; ; i++ ){
		if( fgets( linea, 80, arch ) == NULL ) break;
		ptr = linea;
		for( j=0; j<5; j++ ) {
			k = sscanf( ptr, "%lf", &val );
			if ( k == 1 ) {
				// printf( "%d %d %d\n", i, j, k );
				A[i][j] = val;
				while( *ptr != ' ' && *ptr != '\n' ) ptr++;
				ptr++;
				// printf("%s\n", ptr );
			}
		}
	}
	fclose( arch );
	*mynn = i;
	return( 0 );
}

/**
	Imprime una matrix A de tamao n x n
**/
void imprime( double **A, int n )
{
	int i, j;

	for( i=0; i<n; i++ ) {
		for( j=0; j<n; j++ ) {
			printf( "%lf ", A[i][j] );
		}
		printf("\n");
	}
	printf("\n");
}

int svdLapack( double datos[][5], int ndatos )
{
	double *A, *U;
	double Work[10000];
	double VT[144];
	double S[12];
	double x, y;
	int m, n;
	int LDA, LDU, LDVT, LWORK, INFO;
	char JOBU, JOBVT;
	int itera, k, i, cual;
	struct timeval  tv1, tv2;

	JOBU = 'A';
	JOBVT = 'A';
	m = ndatos*2;
	n = 12;
	LDA = m;
	LDU = m;
	LDVT = n;
	LWORK = 10000;
	// LWORK >= MAX(1,3*MIN(M,N)+MAX(M,N),5*MIN(M,N))
	// max( 1, 3*n+m, 5*m )
	// 10,     100
	// 100,   1000
	// 1000, 10000
	

	A = (double *)malloc( m*n*sizeof(double) );
	if( A==NULL ) {
		fprintf( stderr, "ERROR no hay memoria para la matriz A\n" );
		return( 1 );
	}
				
	U = (double *)malloc( m*m*sizeof(double) );
	if( U==NULL ) {
		free( A );
		fprintf( stderr, "ERROR no hay memoria para la matriz M3\n" );
		return( 2 );
	}

	/**
	  SVD calculado con Lapack
	**/
	gettimeofday(&tv1, NULL);
  for( itera=0; itera<1000; itera++ ) {
  // for( itera=0; itera<1e5; itera++ ) {
  // for( itera=0; itera<1; itera++ ) {

	for( k=0; k<ndatos; k++ ) {

		// printf("%lf %lf %lf %lf %lf\n", datos[k][0], datos[k][1], datos[k][2], datos[k][3], datos[k][4] );
		x = datos[k][3];
		y = datos[k][4];

		// C: 0,0  0,1,  0,2
		// index = n*i + j

		// Fortran:
		//   index = m*j + i
		i = 2*k;

		// (i,j) ->  i + m*j
		A[i] = 0.0;
		A[i+m] = 0.0;
		A[i+2*m] = 0.0;
		A[i+3*m] = 0.0;
		A[i+4*m] = -datos[k][0];
		A[i+5*m] = -datos[k][1];
		A[i+6*m] = -datos[k][2];
		A[i+7*m] = -1.0;
		A[i+8*m] =  datos[k][0]*y;
		A[i+9*m] =  datos[k][1]*y;
		A[i+10*m] = datos[k][2]*y;
		A[i+11*m] = y;

		i++;

		A[i] =  datos[k][0];
		A[i+m] =  datos[k][1];
		A[i+2*m] =  datos[k][2];
		A[i+3*m] = 1.0;
		A[i+4*m] = 0.0;
		A[i+5*m] = 0.0;
		A[i+6*m] = 0.0;
		A[i+7*m] = 0.0;
		A[i+8*m] = -datos[k][0]*x;
		A[i+9*m] = -datos[k][1]*x;
		A[i+10*m] = -datos[k][2]*x;
		A[i+11*m] = -x;

	}

	/**	
	for( i=0; i<ndatos*2; i++ ) {
		for( k=0; k<12; k++ ) 
			printf("%lf ", A[i+k*m] );
		printf("\n");
	}	
	**/


	dgesvd_( &JOBU, &JOBVT, &m, &n, A, &LDA, S, U,
	         &LDU, VT, &LDVT, Work, &LWORK, &INFO);

	// Selecciona el ms pequeo que sea mayor a 1e-8
	cual = 0;
	for( i=11; i>0; i-- ) {
		if( fabs( S[i] ) > 1e-8 ){
			cual = i;
			break;
		}
	}

  }

	gettimeofday(&tv2, NULL);
	printf ("Total time = %f seconds\n",
	     (double) (tv2.tv_usec - tv1.tv_usec) / 1000000 +
	     (double) (tv2.tv_sec - tv1.tv_sec));
	/**
	printf("\nVT:\n");
	for( i=0; i<12; i++ ) {
		for( k=0; k<12; k++ ) 
			printf("%lf ", VT[i+12*k] );
		printf("\n");
	}	
	printf("\n");
	**/

    printf("INFO=%d\n", INFO );
	// for( i=0; i<12; i++ ) 
	// 	printf( "e: %g\n", S[i] );

	printf( "e %d: %g\n", cual, S[cual] );
	printf("rnorm: ");
	for( k=0; k<12; k++ ) {
		printf ("%g ", VT[cual + 12*k] );
	}
	printf("\n");

	free( A );
	free( U );

	return 0;
}

int main( int argc, char **argv )
{
	char nombre[32];
	double datos[1000][5];
	int n;

	if( argc != 2 ) {
		fprintf( stderr, "Args: archivo\n" );
		return( 1 );
	}

	strncpy( nombre, argv[1], 32 );

	if ( lee_archivo( nombre, datos, &n ) != 0 ) {
		fprintf( stderr, "ERROR leyendo archivo de entrada\n" );
		return( 2 );
	}
	printf( "Lei %d datos\n", n );

	svdLapack( datos, n );

	return( 0 );
}
